/*
 *  Dr Geo an interactive geometry software
 * (C) Copyright Hilaire Fernandes  1997-2001
 * hilaire@ofset.org 
 * 
 * This code is copyright Laurent Gauthier 1999
 * lolo@seul.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public Licences as by published
 * by the Free Software Foundation; either version 2; or (at your option)
 * any later version
 *
 * This program is distributed in the hope that it will entertaining,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILTY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Publis License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>

#include "drgeo_tool.h"



/* Helper function */
static void
removeTip (drgeoDrawableUI *drawable)
{
	drawable->startCleanRendering ();
	drawable->refresh ();
	drawable->updateRendering ();
	drawable->endCleanRendering ();
}

// -------------------------------------------------
// Implementation for the Null tool, this tool does nothing.

void drgeoNullTool::
initialize (drgeoDrawableUI * drawable)
{
	// Do nothing.
	printf ("initialize()\n");
}

void drgeoNullTool::
handlePress (drgeoDrawableUI * drawable,
	     drgeoPoint & where)
{
	// Do nothing.
	printf ("handlePress (%lf,%lf)\n", where.getX (), where.getY ());
}

void drgeoNullTool::
handleRelease (drgeoDrawableUI * drawable,
	       drgeoPoint & where)
{
	// Do nothing.
	printf ("handleRelease (%lf,%lf)\n", where.getX (), where.getY ());
}

void drgeoNullTool::
handleMouseAt (drgeoDrawableUI * drawable,
	       drgeoPoint & where)
{
	// Do nothing.
	printf ("handleMouseAt (%lf,%lf)\n", where.getX (), where.getY ());
}

void drgeoNullTool::
handleChoice (drgeoDrawableUI * drawable,
	      geometricObject * item)
{
	// Do nothing.
	printf ("HandleChoice()\n");
}

void drgeoNullTool::
finish (drgeoDrawableUI * drawable)
{
	// Do nothing.
	printf ("finish()\n");
}

// -------------------------------------------------
// Implementation for the Select tool.

void drgeoSelectTool::
initialize (drgeoDrawableUI * drawable)
{
	// This code is invoked when the tool selected.
	state = drgeoStateNeutral;
	tipOn = FALSE;
	last_nb_elem = 0;
}

void drgeoSelectTool::
handlePress (drgeoDrawableUI * drawable,
	     drgeoPoint & where)
{
	liste_elem *list;
	int i;

	// Do nothing if no figure is currently associated with this drawable.
	if (drawable->figure != NULL) {
		if (tipOn) {
			removeTip (drawable);
			tipOn = FALSE;
		}
		switch (state) {
		case drgeoStateItemChoice:
			// The user has not chosen any item, so revert to neutral
			// state.
			state = drgeoStateNeutral;
			// Fall through!  This is really what is wanted here.
		case drgeoStateNeutral:
			// The mouse button is being pressed, check if there are items
			// near the mouse.
			list = drawable->figure->mouseSelection (where);
			list->init_lire ();
			if (list->nb_elem == 1) {
				// Only one item is selected.
				state = drgeoStateItemGrabbed;
				handleChoice (drawable, (geometricObject *) list->lire (0));
				last = where;
			} else if (list->nb_elem > 1) {
				// Several items are selected at once, we must prompt the
				// user for a choice.  When the user is done with his
				// choice the handleChoice method will be invoked.
				state = drgeoStateItemChoice;
				last = where;
				drawable->chooseItem (list);
			}
			else
			{
				// The user click on the background:
				// clear the selection
				state = drgeoStateNeutral;
				drawable->getFigure ()->clearSelection ();
				itemUnderMouse = NULL;
			}
			break;
		case drgeoStateItemSelected:
			// The mouse button is being pressed, check if the last
			// selected item is near the mouse.
			list = drawable->figure->mouseSelection (where);
			list->init_lire ();
			if (list->nb_elem >= 1) {
				// Check each item under the mouse to see if the last item
				// selected is there.  If so select it, if not go back to
				// neutral state.
				for (i = 0; i < list->nb_elem; i++) {
					if (this->item == (geometricObject *) list->lire (0)) {
						state = drgeoStateItemGrabbed;
						// The following is not necessary because it has
						// already been performed, when the item has been
						// selected among multiple items.
						// handleChoice (drawable, this->item);
						last = where;
						break;	// Get out of the loop!  We found it.

					}
				}
				if (state == drgeoStateItemSelected) {
					// The item last selected is not under the mouse at
					// this point, so we are back to neutral state.
					state = drgeoStateNeutral;
				}
			} else {
				// No item under the mouse go back to the Neutral state.
				state = drgeoStateNeutral;
			}
			break;
		default:
			// If we get there, then there's a problem with the event
			// handling code... XXX
			break;
		}
	}
}

void drgeoSelectTool::
handleRelease (drgeoDrawableUI * drawable,
	       drgeoPoint & where)
{
	// Do nothing if no figure is currently associated with this drawable.
	if (drawable->figure != NULL) {
		liste_elem *list;
		geometricObject *item;

		if (tipOn) {
			removeTip (drawable);
			tipOn = FALSE;
		}
		switch (state) {
		case drgeoStateItemChoice:
		case drgeoStateItemSelected:
			// The user has not chosen any item, so revert to neutral
			// state.
			state = drgeoStateNeutral;
			// Fall through!  This is really what is wanted here.
		case drgeoStateNeutral:
			// The mouse button is just being released with nothing being
			// dragged, so we only show a tip displaying the full name of
			// the nearest item.
			list = drawable->figure->mouseSelection (where);
			list->init_lire ();
			if (list->nb_elem == 1) {
				item = (geometricObject *) list->lire (0);
				drawable->showTip (where, item->getTypeName ());
				tipOn = TRUE;
			} else if (list->nb_elem > 1) {
				// There are several items under the mouse.
				drawable->showTip (where, "???");
				tipOn = TRUE;
			}
			break;
		case drgeoStateItemGrabbed:
			// The user is just releasing the mouse.  Reset the state to
			// neutral and that's all.
			state = drgeoStateNeutral;
			break;
		case drgeoStateItemDragged:
			// The drag is finished, so we just drop the selected ietms
			// where the mouse currently is.
			state = drgeoStateNeutral;
			drawable->figure->moveItem (NULL, where - start);
			drawable->figure->dropSelection (last, where);
			/* record the drag path */
			drawable->endCleanRendering ();
			removeTip (drawable);

			// Forget the item when dragged
			drawable->getFigure ()->clearSelection ();
			itemUnderMouse = NULL;

			break;
		default:
			// If we get there, then there's a problem with the event
			// handling code... XXX
			break;
		}
	}
}

void drgeoSelectTool::
handleMouseAt (drgeoDrawableUI * drawable,
	       drgeoPoint & where)
{
	int i;

	// Do nothing if no figure is currently associated with this drawable.
	if (drawable->figure != NULL) {
		liste_elem *list;

		switch (state) {
		case drgeoStateItemChoice:
			// The user has not chosen any item, so revert to neutral
			// state.
			state = drgeoStateNeutral;
			// Fall through!  This is really what is wanted here.
		case drgeoStateNeutral:
			// The mouse is just moving over the figure, so we just query
			// the figure to know which items are found near the current
			// mouse position.  And a tooltip is shown displaying the full
			// name of the nearest item.
			list = drawable->figure->mouseSelection (where);
			if (list->nb_elem == 1 && 
			    ((geometricObject *) list->lire (1) != itemUnderMouse) ) {
				// The mouse move over another item
				if (tipOn)
					removeTip (drawable);
				itemUnderMouse = (geometricObject *) list->lire (1);
				drawable->showTip (where, itemUnderMouse->getTypeName ());
				tipOn = TRUE;
				drawable->setCursor (hand);
			} else if (list->nb_elem > 1 && last_nb_elem <= 1) {
				// There are NOW several items under the mouse.
				if (tipOn)
					removeTip (drawable);
				drawable->showTip (where, "???");
				tipOn = TRUE;
				drawable->setCursor (hand);
				itemUnderMouse = NULL;
			} 
			else if (list->nb_elem == 0 && tipOn)
			{
				/* mouse not over object anymore, remove the tip */	
				removeTip (drawable);
				drawable->setCursor (arrow);
				tipOn = FALSE;
				itemUnderMouse = NULL;				
			}
			last_nb_elem = list->nb_elem;
			break;
		case drgeoStateItemSelected:
			// The mouse is moving over the figure, check if we are moving
			// away from the spot where the last item was selected.  If so
			// go back to neutral state.
			if (away
			    && (spot - where).norm () > (last - spot).norm () 
			    + 4.0 * drawable->getRange ()) 
			{
				state = drgeoStateNeutral;
				drawable->figure->clearSelection ();
				drawable->setCursor (arrow);																
			} else {
				// If the last selected item is under the mouse, then show
				// a tip about it.
				list = drawable->figure->mouseSelection (where);
				list->init_lire ();
				if (list->nb_elem >= 1) {
					// Check each item under the mouse to see if the last item
					// selected is there.  If so show its tip.
					for (i = 0; i < list->nb_elem; i++) {
						if (this->item == (geometricObject *) list->lire (0) 
						    && !tipOn)
						{
							drawable->showTip (where, this->item->getTypeName ());
							tipOn = TRUE;
							drawable->setCursor (hand);																
							break;	// Get out of the loop!  We found it.
						}
					}
				}
				away = TRUE;
				last = where;
			}
			break;
		case drgeoStateItemGrabbed:
			// The user is just starting a mouse move with an item
			// grabbed, this means a drag is starting.  Notify the
			// figure.
			// Prepare the double buffering
			drawable->startCleanRendering ();
			state = drgeoStateItemDragged;
			drawable->figure->dragSelection (last, where);
			drawable->refresh ();
			last = where;
			start = where;
			break;
		case drgeoStateItemDragged:
			// The drag is going on.
			drawable->updateRendering ();
			drawable->figure->dragSelection (last, where);
			drawable->refresh ();
			last = where;
			break;
		default:
			// If we get there, then there's a problem with the event
			// handling code... XXX
			break;
		}
	}
}

void drgeoSelectTool::
handleChoice (drgeoDrawableUI * drawable,
	      geometricObject * item)
{
	drgeoPoint xxx;

	if (drawable->figure) {
		if (state == drgeoStateItemChoice) {
			state = drgeoStateItemSelected;
			this->item = item;	// Keep track of the last item
			// selected for later reference.

			away = FALSE;
			spot = last;
		}
		drawable->figure->clearSelection ();
		drawable->figure->addToSelection (xxx, item);
	}
}

void drgeoSelectTool::
finish (drgeoDrawableUI * drawable)
{
	// This code is invoked when the tool is deselected.
}

// -------------------------------------------------
// Implementation for the Build tool.

void drgeoBuildTool::
initialize (drgeoDrawableUI * drawable)
{
	// This code is invoked when the tool selected.
	state = drgeoStateNeutral;
	tipOn = FALSE;
	last_nb_elem = 0;
}

void drgeoBuildTool::
handlePress (drgeoDrawableUI * drawable,
	     drgeoPoint & where)
{
	liste_elem *list;
	int i;

	// Do nothing if no figure is currently associated with this drawable.
	if (drawable->figure != NULL) {
		if (tipOn) {
			removeTip (drawable);
			tipOn = FALSE;
		}
		switch (state) {
		case drgeoStateItemChoice:
			// The user has not chosen any item, so revert to neutral
			// state.
			state = drgeoStateNeutral;
			// Fall through!  This is really what is wanted here.
		case drgeoStateNeutral:
			// The mouse button is being pressed, check if there are items
			// near the mouse.
			list = drawable->figure->mouseSelection (where);
			list->init_lire ();
			if (list->nb_elem == 0) {
				// no item was selected.  Just invoke handleChoice with a
				// NULL parameter that will be transmitted to
				// figure->addToSelection() for appropriate processing.
				last = where;
				handleChoice (drawable, NULL);
			} else if (list->nb_elem == 1) {
				// Only one item is selected, add it to the selection.
				last = where;
				handleChoice (drawable, (geometricObject *) list->lire (0));
			} else if (list->nb_elem > 1) {
				// Several items are selected at once, we must prompt the
				// user for a choice.  When the user is done with his
				// choice the handleChoice method will be invoked.
				state = drgeoStateItemChoice;
				last = where;
				drawable->chooseItem (list);
			}
			break;
		default:
			// If we get there, then there's a problem with the event
			// handling code... XXX
			break;
		}
	}
}

void drgeoBuildTool::
handleRelease (drgeoDrawableUI * drawable,
	       drgeoPoint & where)
{
	// Do nothing if no figure is currently associated with this drawable.
	if (drawable->figure != NULL) {
		liste_elem *list;
		geometricObject *item;

		if (tipOn) {
			removeTip (drawable);
			tipOn = FALSE;
		}
		switch (state) {
		case drgeoStateItemChoice:
			// The user has not chosen any item, so revert to neutral
			// state.
			state = drgeoStateNeutral;
			// Fall through!  This is really what is wanted here.
		case drgeoStateNeutral:
			// The mouse button is just being released with nothing being
			// dragged, so we only show a tip displaying the full name of
			// the nearest item.
			list = drawable->figure->mouseSelection (where);
			list->init_lire ();
			if (list->nb_elem == 1) {
				item = (geometricObject *) list->lire (0);
				drawable->showTip (where, item->getTypeName ());
				tipOn = TRUE;
			} else if (list->nb_elem > 1) {
				// There are several items under the mouse.
				drawable->showTip (where, "???");
				tipOn = TRUE;
			}
			break;
		default:
			// If we get there, then there's a problem with the event
			// handling code... XXX
			break;
		}
	}
}

void drgeoBuildTool::
handleMouseAt (drgeoDrawableUI * drawable,
	       drgeoPoint & where)
{
	int i;

	// Do nothing if no figure is currently associated with this drawable.
	if (drawable->figure != NULL) {
		liste_elem *list;
		geometricObject *item;

		switch (state) {
		case drgeoStateItemChoice:
			// The user has not chosen any item, so revert to neutral
			// state.
			state = drgeoStateNeutral;
			// Fall through!  This is really what is wanted here.
		case drgeoStateNeutral:
			// The mouse is just moving over the figure, so we just query
			// the figure to know which items are found near the current
			// mouse position.  And a tooltip is shown displaying the full
			// name of the nearest item.
			list = drawable->figure->mouseSelection (where);
			if (list->nb_elem == 1 &&
			    ((geometricObject *) list->lire (1) != itemUnderMouse) )
			{
				if (tipOn)
					removeTip (drawable);
				itemUnderMouse = (geometricObject *) list->lire (1);
				drawable->showTip (where, itemUnderMouse->getTypeName ());
				tipOn = TRUE;
				drawable->setCursor (pen);
			} else if (list->nb_elem > 1 && last_nb_elem <= 1) {
				// There are several items under the mouse.
				if (tipOn)
					removeTip (drawable);
				drawable->showTip (where, "???");
				tipOn = TRUE;
				drawable->setCursor (pen);
				itemUnderMouse = NULL;
			}
			else if (list->nb_elem == 0 &&  tipOn)
			{
				/* mouse not over object anymore, remove the tip */
				removeTip (drawable);
				drawable->setCursor (arrow);
				tipOn = FALSE;
				itemUnderMouse = NULL;
			}
			last_nb_elem = list->nb_elem;
			break;
		default:
			// If we get there, then there's a problem with the event
			// handling code... XXX
			break;
		}
	}
}

void drgeoBuildTool::
handleChoice (drgeoDrawableUI * drawable,
	      geometricObject * item)
{
	if (drawable->figure) {
		// Add the selected item to the current figure's selection.
		drawable->figure->addToSelection (last, item);
	}
}

void drgeoBuildTool::
finish (drgeoDrawableUI * drawable)
{
	// This code is invoked when the tool is deselected.
}



// -----------------------------------------------------
// Macro Build Tool, used for macro creation.
// -----------------------------------------------------

void drgeoMacroBuildTool::
initialize (drgeoDrawableUI * drawable)
{
	// First invoke the initialize() method of drgeoBuildTool
	// super-class.
      drgeoBuildTool::initialize (drawable);

	// Now let's do our stuff.
	builder = new drgeoMacroBuilder (drawable->getFigure ());

	// Create the macro dialog that will control the builder.
	dialog = drawable->createMacroBuildDialog (builder);
}

// All the other methods are inherited from drgeoBuildTool.  We only
// redefine the handleChoice() method, and of course the initialize()
// and finish().

void drgeoMacroBuildTool::
handleChoice (drgeoDrawableUI * drawable,
	      geometricObject * item)
{
	// Forward this item to the macro builder.
	if (dialog && item) {
		dialog->add (item);
	}
}

void drgeoMacroBuildTool::
finish (drgeoDrawableUI * drawable)
{
	// First invoke the initialize() method of drgeoBuildTool
	// super-class.
      drgeoBuildTool::finish (drawable);

	// Remove the macro dialog that is controlling the builder.
	delete dialog;

	// Free the drgeoMacroBuilder.
	delete builder;
}

// -----------------------------------------------------
// Macro Play Tool, used for macro execution.
// -----------------------------------------------------

void drgeoMacroPlayTool::
initialize (drgeoDrawableUI * drawable)
{
	// First invoke the initialize() method of drgeoBuildTool
	// super-class.
      drgeoBuildTool::initialize (drawable);

	// Now let's do our stuff.
	player = new drgeoMacroPlayer (drawable->getFigure ());

	// Create the macro dialog that will control the Player.
	dialog = drawable->createMacroPlayDialog (player);
}

// All the other methods are inherited from drgeoBuildTool.  We only
// redefine the handleChoice() method, and of course the initialize()
// and finish().

void drgeoMacroPlayTool::
handleChoice (drgeoDrawableUI * drawable,
	      geometricObject * item)
{
	drgeoPoint xxx;

	// Forward this item to the macro Player.
	if (dialog && item) {
		dialog->add (item);
	}
}

void drgeoMacroPlayTool::
finish (drgeoDrawableUI * drawable)
{
	// First invoke the initialize() method of drgeoPlayTool
	// super-class.
      drgeoBuildTool::finish (drawable);

	// Remove the macro dialog that is controlling the Player.
	delete dialog;

	// Free the drgeoMacroPlayer.
	delete player;
}

// -----------------------------------------------------
// Delete Tool, used for object deletion.
// -----------------------------------------------------

void drgeoDeleteTool::
handleChoice (drgeoDrawableUI * drawable,
	      geometricObject * item)
{
	char *message;

	// Forward this item to the macro builder.
	if (drawable->figure && item != NULL) {
		// Hide the removed items just to show the consequence of the
		// removal.
		drawable->figure->hideRemovedItems (item);

		// Ask the user if he wants to proceed with the item deletion.
		if (item->getName ())
		{
			if (strlen (item->getName ())) 
				message = g_strdup_printf (N_ ("Remove %s?"), item->getName ());
		}
		else 
			message = g_strdup (N_ ("Remove selected item?"));		
		if (drawable->askOkCancel (message)) {
			// Remove the item if the user wants it.
			drawable->figure->removeItem (item);
		}
		// Show again the list of items to be removed.
		drawable->figure->showRemovedItems ();
		drawable->figure->redraw (FALSE);
	}
}






// -----------------------------------------------------
// Style Editor Tool, used to adjust colors, ...
// -----------------------------------------------------

void drgeoStyleTool::
initialize (drgeoDrawableUI * drawable)
{
	// First invoke the initialize() method of drgeoBuildTool
	// super-class.
      drgeoBuildTool::initialize (drawable);

	dialog = drawable->createStyleDialog ();
}

// All the other methods are inherited from drgeoBuildTool.  We only
// redefine the handleChoice() method, and of course the initialize()
// and finish().

void drgeoStyleTool::
handleChoice (drgeoDrawableUI * drawable,
	      geometricObject * item)
{
	drgeoPoint xxx;
	// Forward this item to the style dialog.
	if (dialog && item) {
		// Update the dialog that match the item nwly selected.
		drawable->figure->clearSelection ();
		drawable->figure->addToSelection (xxx, item);
		dialog->edit (item);
	}
}

void drgeoStyleTool::
finish (drgeoDrawableUI * drawable)
{
	// First invoke the initialize() method of drgeoBuildTool
	// super-class.
      drgeoBuildTool::finish (drawable);

	// Remove the macro dialog that is used to adjust style.
	if (dialog) {
		delete dialog;
	}
}

// -----------------------------------------------------
// Property Editor Tool, used to adjust properties when
// availables.
// -----------------------------------------------------

void drgeoPropertyTool::
initialize (drgeoDrawableUI * drawable)
{
	// First invoke the initialize() method of drgeoBuildTool
	// super-class.
      drgeoBuildTool::initialize (drawable);

	dialog = drawable->createPropertyDialog ();
}

void drgeoPropertyTool::
handleChoice (drgeoDrawableUI * drawable,
	      geometricObject * item)
{
	gint nb;
	drgeoPoint xxx;
	if (drawable->figure) 
	{
		// Add the selected item to the current figure's selection.
		drawable->figure->clearSelection ();
		nb = drawable->figure->addToSelection (xxx, item);
	}
	// Forward this item to the style dialog.
	if (dialog && item && nb > 0) {
		// Update the dialog that match the item newly selected.
		dialog->edit (item);
	}
}

void drgeoPropertyTool::
finish (drgeoDrawableUI * drawable)
{
	// First invoke the initialize() method of drgeoBuildTool
	// super-class.
      drgeoBuildTool::finish (drawable);

      drawable->figure->clearSelection ();
	// Remove the macro dialog that is used to adjust style.
	if (dialog) {
		delete dialog;
	}
}




